import collections
import inspect
import logging
import logging.config
import os
import threading

import structlog
from structlog import stdlib
from pydantic import BaseModel


PROD_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters":
        {
            "json": {
                "format": "%(message)s %(lineno)d %(pathname)s",
                "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
            }
        },
    "handlers": {
        "json": {
            "class": "logging.StreamHandler",
            "formatter": "json"
        }
        },
    "loggers": {
        "": {
            "handlers": ["json"],
            "level": "INFO"
        }
        },
}


DEV_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "default": {
            "level": "DEBUG",
            "()": "logging.StreamHandler",
        }
    },
    "loggers": {
        "": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": True
        }
    },
}


def is_dev():
    return "PROD_ENV" not in os.environ


default_log_level = "DEBUG" if is_dev() else "INFO"
log_level = os.environ.get("LOG_LEVEL", default_log_level)
LOGGING_CONFIG = DEV_LOGGING if is_dev() else PROD_LOGGING


def get_logger(logger_name):
    if logger_name is None:
        logger_name = inspect.currentframe().f_back.f_globals["__name__"]
    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)
    return structlog.wrap_logger(logger)


def _add_thread_info(logger, method_name, event_dict):  # pylint: disable=unused-argument
    thread = threading.current_thread()
    event_dict["thread_id"] = thread.ident
    event_dict["thread_name"] = thread.name
    return event_dict


def _render_models(logger, method_name, event_dict):
    return {k: v.dict() if isinstance(v, BaseModel) else v for k, v in event_dict.items()}


def _order_keys(logger, method_name, event_dict):  # pylint: disable=unused-argument
    return collections.OrderedDict(sorted(event_dict.items(), key=lambda item: (item[0] != "event", item)))


logging.config.dictConfig(LOGGING_CONFIG)
processors_list = [
    # This performs the initial filtering, so we don't
    # evaluate e.g. DEBUG when unnecessary
    stdlib.filter_by_level,
    _add_thread_info,
    # Adds logger=module_name (e.g __main__)
    # Adds level=info, debug, etc.
    structlog.stdlib.add_log_level,
    # Performs the % string interpolation as expected
    structlog.stdlib.PositionalArgumentsFormatter(True),
    # Include the stack when stack_info=True
    structlog.processors.StackInfoRenderer(),
    # Include the exception when exc_info=True
    # e.g log.exception() or log.warning(exc_info=True)'s behavior
    structlog.processors.format_exc_info,
    # add the logger name
    structlog.stdlib.add_logger_name,
]
if not is_dev():
    processors_list = processors_list + [
        _order_keys,
        structlog.processors.JSONRenderer()  # in prod, render to json
    ]
else:
    processors_list = processors_list + [
        _order_keys,
        # renders to console (stdout)
        structlog.dev.ConsoleRenderer(repr_native_str=True)
    ]

if not structlog.is_configured():
    structlog.configure_once(
        context_class=dict,
        logger_factory=stdlib.LoggerFactory(),
        wrapper_class=stdlib.BoundLogger,
        processors=processors_list,
    )

logger = get_logger("SDK")


import re

from typing import Optional, List, Callable, Dict, Any, Union, Tuple

FASTAPI_PATH_VARIABLE = re.compile(r"\{(\w+)\}")
PLACEHOLDER = r"(\w+)"


def extract_pattern_and_context(path: str):
    # we want to match two options: both ending and non ending with /
    if path.endswith("/"):
        path = path[:-1]
    parts = path.split("/")
    context_vars = []
    for i, part in enumerate(parts):
        match = FASTAPI_PATH_VARIABLE.match(part)
        if match:
            context_vars.append(match.group(1))
            parts[i] = PLACEHOLDER
        else:
            parts[i] = re.escape(parts[i])
    path_pattern = "^" + "/".join(parts) + r"[/]?$"
    return path_pattern, context_vars


class ActionDefinition:
    def __init__(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        path: Optional[str] = None,
        resource_id: Optional[str] = None,
        attributes: Dict[str, Any] = {},
    ):
        self.name = name
        self.title = title
        self.description = description
        self.path = path
        self.attributes = attributes
        self._resource_id = resource_id
        self._resource_name = None

    @property
    def resource_id(self) -> str:
        return self._resource_id

    @property
    def resource_name(self) -> str:
        return self._resource_name

    def set_resource_id(self, resource_id: str):
        self._resource_id = resource_id

    def set_resource_name(self, resource_name: str):
        self._resource_name = resource_name

    def dict(self):
        d = {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "path": self.path,
            "attributes": self.attributes,
        }
        if self._resource_id is not None:
            d["resource_id"] = self._resource_id
        return d

    def __repr__(self):
        return "ActionDefinition(name='{}', path='{}')".format(self.name, self.path)


class ResourceDefinition:
    def __init__(
        self,
        name: str,
        type: str,
        path: str,
        description: str = None,
        actions: List[ActionDefinition] = [],
        attributes: Dict[str, Any] = {},
    ):
        self.name = name
        self.type = type
        self.path = path
        self.description = description
        self.actions = actions
        self.attributes = attributes
        self._remote_id = None

    @property
    def remote_id(self) -> str:
        return self._remote_id

    def set_remote_id(self, remote_id: str):
        self._remote_id = remote_id

    def dict(self):
        return {
            "name": self.name,
            "type": self.type,
            "path": self.path,
            "description": self.description,
            "actions": [a.dict() for a in self.actions],
            "attributes": self.attributes,
        }

    def __repr__(self):
        return "ResourceDefinition(name='{}', path='{}', actions={})".format(
            self.name, self.path, repr(self.actions)
        )


class ResourceRegistry:
    def __init__(self):
        self._resources = {}
        self._already_synced = set()
        self._processed_paths = set()
        self._path_regexes = []

    @property
    def resources(self):
        return self._resources.values()

    def add_resource(self, resource: ResourceDefinition):
        if not resource.name in self._resources:
            self._resources[resource.name] = resource
            self._process_path(resource.path, resource.name)

        for action in resource.actions:
            action.set_resource_name(resource.name)
            self._process_path(action.path, resource.name)

    def add_action_to_resource(
        self, resource_name: str, action: ActionDefinition
    ) -> Optional[ActionDefinition]:
        if not resource_name in self._resources:
            return None

        resource = self._resources[resource_name]
        action.set_resource_id(resource.remote_id)
        action.set_resource_name(resource.name)

        existing_actions = [action.name for action in resource.actions]
        if not action.name in existing_actions:
            resource.actions.append(action)

        self._process_path(action.path, resource.name)
        return action

    def is_synced(self, obj: Union[ResourceDefinition, ActionDefinition]) -> bool:
        if isinstance(obj, ResourceDefinition):
            return obj.name in self._already_synced

        if isinstance(obj, ActionDefinition):
            return self.action_key(obj) in self._already_synced

        return False

    def mark_as_synced(
        self, obj: Union[ResourceDefinition, ActionDefinition], remote_id: str
    ):
        if isinstance(obj, ResourceDefinition):
            self._already_synced.add(obj.name)
            self._resources[obj.name].set_remote_id(remote_id)
            for action in obj.actions:
                self._already_synced.add(self.action_key(action))

        if isinstance(obj, ActionDefinition):
            self._already_synced.add(self.action_key(obj))

    @classmethod
    def action_key(cls, action: ActionDefinition) -> str:
        return "{}:{}".format(action.resource_name, action.name)

    def _process_path(self, path: str, resource_name: str):
        if path in self._processed_paths:
            return

        path_pattern, context_vars = extract_pattern_and_context(path)
        self._path_regexes.append(
            dict(pattern=path_pattern, context=context_vars, resource_name=resource_name)
        )
        self._processed_paths.add(path)

    def get_resource_by_path(self, path: str) -> Tuple[Optional[str], Optional[ResourceDefinition]]:
        for potential in self._path_regexes:
            pattern = potential["pattern"]
            resource_name = potential["resource_name"]
            context_vars = potential["context"]
            m = re.match(pattern, path)
            if m:
                resource_def = self._resources.get(resource_name, None)
                match_groups = m.groups()
                if len(context_vars) == len(match_groups):
                    context = dict(zip(context_vars, match_groups))
                else:
                    context = {}
                return resource_name, resource_def, context
        return None, None, {}

resource_registry = ResourceRegistry()



from typing import Optional, List, Callable, Dict, Any

from .client import authorization_client, ResourceStub
from .resource_registry import ResourceDefinition, ActionDefinition
from .enforcer import enforcer
from .markers import resource_id, resource_type, org_id
from .constants import SIDECAR_URL
from .logger import logger

def init(token, app_name, service_name, **kwargs):
    """
    inits the authorizon client
    """
    logger.info(f"authorizon.init", sidecar_url=SIDECAR_URL)
    authorization_client.initialize(
        token=token, app_name=app_name, service_name=service_name, **kwargs
    )

def resource(
    name: str,
    type: str,
    path: str,
    description: str = None,
    actions: List[ActionDefinition] = [],
    attributes: Optional[Dict[str, Any]] = None,
    **kwargs
) -> ResourceStub:
    """
    declare a resource type.

    usage:

    authorizon.resource(
        name="Todo",
        description="todo item",
        type=authorizon.types.REST,
        path="/lists/{list_id}/todos/{todo_id}",
        actions=[
            authorizon.action(
                name="add",
                title="Add",
                path="/lists/{list_id}/todos/",
                verb="post",
            ),
            ...
        ]
    )

    you can later add actions on that resource:

    todo = authorizon.resource( ... )
    todo.action(
        name="add",
        title="Add",
        path="/lists/{list_id}/todos/",
        verb="post",
    )
    """
    attributes = attributes or {}
    attributes.update(kwargs)
    resource = ResourceDefinition(
        name=name,
        type=type,
        path=path,
        description=description,
        actions=actions,
        attributes=attributes
    )
    return authorization_client.add_resource(resource)

def action(
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        path: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ActionDefinition:
    """
    declare an action on a resource.

    usage:
    authorizon.resource(
        ...,
        actions = [
            authorizon.action(...),
            authorizon.action(...),
        ]
    )
    """
    attributes = attributes or {}
    attributes.update(kwargs)
    return ActionDefinition(
        name=name,
        title=title,
        description=description,
        path=path,
        attributes=attributes
    )

sync_user = authorization_client.sync_user
delete_user = authorization_client.delete_user
sync_org = authorization_client.sync_org
delete_org = authorization_client.delete_org
add_user_to_org = authorization_client.add_user_to_org
remove_user_from_org = authorization_client.remove_user_from_org
get_orgs_for_user = authorization_client.get_orgs_for_user
assign_role = authorization_client.assign_role
unassign_role = authorization_client.unassign_role
update_policy_data = authorization_client.update_policy_data

is_allowed = enforcer.is_allowed
transform_resource_context = enforcer.add_transform
provide_context = enforcer.add_context


from .api import (
    init,
    resource,
    action,
    sync_user,
    delete_user,
    sync_org,
    delete_org,
    add_user_to_org,
    remove_user_from_org,
    get_orgs_for_user,
    assign_role,
    unassign_role,
    is_allowed,
    resource_id,
    resource_type,
    org_id,
    transform_resource_context,
    provide_context,
    update_policy_data,
)

__all__ = [
    'init',
    'resource',
    'action',
    'sync_user',
    'delete_user',
    'sync_org',
    'delete_org',
    'add_user_to_org',
    'remove_user_from_org',
    'get_orgs_for_user',
    'assign_role',
    'unassign_role',
    'set_user',
    'set_org',
    'set_context',
    'is_allowed',
    'resource_id',
    'resource_type',
    'org_id',
    'transform_resource_context',
    'provide_context',
    'update_policy_data',
]


import requests
import json

from typing import Optional, Dict, Any

from .constants import SIDECAR_URL
from .resource_registry import resource_registry, ResourceDefinition, ActionDefinition
from .logger import logger

class ResourceStub:
    def __init__(self, resource_name: str):
        self._resource_name = resource_name

    def action(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        path: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        attributes = attributes or {}
        attributes.update(kwargs)
        action = ActionDefinition(
            name=name,
            title=title,
            description=description,
            path=path,
            attributes=attributes
        )
        authorization_client.add_action_to_resource(self._resource_name, action)


class AuthorizationClient:
    def __init__(self):
        self._initialized = False
        self._registry = resource_registry

    def initialize(self, token, app_name, service_name, **kwargs):
        self._token = token
        self._client_context = {"app_name": app_name, "service_name": service_name}
        self._client_context.update(kwargs)
        self._initialized = True
        self._requests = requests.session()
        self._requests.headers.update(
            {"Authorization": "Bearer {}".format(self._token)}
        )
        self._sync_resources()

    @property
    def token(self):
        self._throw_if_not_initialized()
        return self._token

    def update_policy(self):
        self._throw_if_not_initialized()
        self._requests.post(f"{SIDECAR_URL}/update_policy")

    def update_policy_data(self):
        self._throw_if_not_initialized()
        self._requests.post(f"{SIDECAR_URL}/update_policy_data")

    def add_resource(self, resource: ResourceDefinition) -> ResourceStub:
        self._registry.add_resource(resource)
        self._maybe_sync_resource(resource)
        return ResourceStub(resource.name)

    def add_action_to_resource(self, resource_name: str, action: ActionDefinition):
        action = self._registry.add_action_to_resource(resource_name, action)
        if action is not None:
            self._maybe_sync_action(action)

    def _maybe_sync_resource(self, resource: ResourceDefinition):
        if self._initialized and not self._registry.is_synced(resource):
            logger.info("syncing resource", resource=repr(resource))
            response = self._requests.put(
                f"{SIDECAR_URL}/sdk/resource",
                data=json.dumps(resource.dict()),
            )
            self._registry.mark_as_synced(
                resource, remote_id=response.json().get('id'))

    def _maybe_sync_action(self, action: ActionDefinition):
        resource_id = action.resource_id
        if resource_id is None:
            return

        if self._initialized and not self._registry.is_synced(action):
            logger.info("syncing action", action=repr(action))
            response = self._requests.put(
                f"{SIDECAR_URL}/sdk/resource/{resource_id}/action",
                data=json.dumps(action.dict())
            )
            self._registry.mark_as_synced(
                action, remote_id=response.json().get('id'))

    def _sync_resources(self):
        # will also sync actions
        for resource in self._registry.resources:
            self._maybe_sync_resource(resource)

    def sync_user(
        self,
        user_id: str,
        user_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        self._throw_if_not_initialized()
        data = {
            "id": user_id,
            "data": user_data
        }
        response = self._requests.put(
            f"{SIDECAR_URL}/sdk/user",
            data=json.dumps(data),
        )
        return response.json()

    def delete_user(
        self,
        user_id: str
    ):
        self._throw_if_not_initialized()
        self._requests.delete(
            f"{SIDECAR_URL}/sdk/user/{user_id}",
        )

    def sync_org(
        self,
        org_id: str,
        org_name: str,
        org_metadata: Dict[str, Any]={}
    ):
        self._throw_if_not_initialized()
        data = {
            "external_id": org_id,
            "name": org_name,
        }
        response = self._requests.post(
            f"{SIDECAR_URL}/sdk/organization",
            data=json.dumps(data),
        )
        return response.json()

    def delete_org(
        self,
        org_id: str
    ):
        self._throw_if_not_initialized()
        self._requests.delete(
            f"{SIDECAR_URL}/sdk/organization/{org_id}",
        )

    def add_user_to_org(
        self,
        user_id: str,
        org_id: str
    ):
        self._throw_if_not_initialized()
        data = {
            "user_id": user_id,
            "org_id": org_id,
        }
        response = self._requests.post(
            f"{SIDECAR_URL}/sdk/add_user_to_org",
            data=json.dumps(data),
        )
        return response.json()

    def remove_user_from_org(
        self,
        user_id: str,
        org_id: str
    ):
        self._throw_if_not_initialized()
        data = {
            "user_id": user_id,
            "org_id": org_id,
        }
        response = self._requests.post(
            f"{SIDECAR_URL}/sdk/remove_user_from_org",
            data=json.dumps(data),
        )
        return response.json()

    def get_orgs_for_user(
        self,
        user_id: str
    ):
        self._throw_if_not_initialized()
        response = self._requests.get(
            f"{SIDECAR_URL}/sdk/get_orgs_for_user/{user_id}",
        )
        return response.json()

    def assign_role(
        self,
        role: str,
        user_id: str,
        org_id: str
    ):
        self._throw_if_not_initialized()
        data = {
            "role": role,
            "user_id": user_id,
            "org_id": org_id,
        }
        response = self._requests.post(
            f"{SIDECAR_URL}/sdk/assign_role",
            data=json.dumps(data),
        )
        return response.json()

    def unassign_role(
        self,
        role: str,
        user_id: str,
        org_id: str
    ):
        self._throw_if_not_initialized()
        data = {
            "role": role,
            "user_id": user_id,
            "org_id": org_id,
        }
        response = self._requests.post(
            f"{SIDECAR_URL}/sdk/unassign_role",
            data=json.dumps(data),
        )
        return response.json()

    def _throw_if_not_initialized(self):
        if not self._initialized:
            raise RuntimeError("You must call authorizon.init() first!")


authorization_client = AuthorizationClient()



from setuptools import setup, find_packages

def get_requirements(env=""):
    if env:
        env = "-{}".format(env)
    with open("requirements{}.txt".format(env)) as fp:
        return [x.strip() for x in fp.read().split("\n") if not x.startswith("#")]

setup(
    name='authorizon',
    version='0.0.4',
    packages=find_packages(),
    author='Or Weis, Asaf Cohen',
    author_email="or@authorizon.com",
    python_requires='>=3.8',
    description='authorizon python sdk',
    install_requires=get_requirements(),
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)


from functools import wraps
from typing import Callable

class AuthorizationMarkers:
    def __init__(self):
        self._registry = {}

    def register_method(self, cls_name, method_name, authz_alias):
        self._registry.setdefault(cls_name, {})[authz_alias] = method_name

    @property
    def registry(self):
        return self._registry

authorization_markers = AuthorizationMarkers()

def mark(marker: str) -> Callable:
    """
    this decorator allows us to mark methods on inspected resource objects,
    so that we can later evaluate the correct properties for enforcement.
    """
    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            authorization_markers.register_method(self.__class__.__name__, f.__name__, marker)
            return f(self, *args, **kwargs)
        return wrapper
    return decorator

resource_id = mark('resource_id')
resource_type = mark('resource_type')
org_id = mark('org_id')


from typing import Optional
from .resource_registry import resource_registry, ResourceDefinition

class ResourceNotFound(Exception):
    pass

class Resource:
    """
    temp resource object
    """
    def __init__(self, name: str, path: str, definition: Optional[ResourceDefinition]=None, context={}):
        self._name = name
        self._instance_path = path
        self._path = definition.path if definition else None
        self._context = context
        self._definition = definition

    def dict(self):
        return {
            "type": self._name,
            "path": self._path,
            "instance": self._instance_path,
            "context": self._context
        }

    def __repr__(self):
        return "authorizon.Resource(name={}, path={}, instance={})".format(
            self._name, self._path, self._instance_path
        )

    @classmethod
    def from_path(self, path: str):
        name, resource_def, context = resource_registry.get_resource_by_path(path)
        if name is None:
            raise ResourceNotFound()
        return Resource(name, path, resource_def, context)


import requests
import json
import copy

from typing import Optional, Dict, Any, Callable, Union

from .constants import SIDECAR_URL
from .resource import Resource, ResourceNotFound
from .logger import logger

def set_if_not_none(d: dict, k: str, v):
    if v is not None:
        d[k] = v

Context = Dict[str, Any]
ContextTransform = Callable[[Context], Context]
ResourceType = Union[str, Resource, Dict[str, Any]]

class Enforcer:
    POLICY_NAME = "rbac"

    def __init__(self):
        self._transforms = []
        self._context = {}

    def add_transform(self, transform: ContextTransform):
        self._transforms.append(transform)

    def add_context(self, context: Context):
        self._context.update(context)

    def _combine_context(self, query_context: Context) -> Context:
        combined_context = {}
        combined_context.update(self._context)
        combined_context.update(query_context)
        return combined_context

    def _transform_context(self, initial_context: Context) -> Context:
        context = copy.deepcopy(initial_context)
        for transform in self._transforms:
            context = transform(context)
        return context

    def _translate_resource(self, resource: ResourceType) -> Dict[str, Any]:
        resource_dict = {}
        if isinstance(resource, str):
            try:
                resource_dict = Resource.from_path(resource).dict()
            except ResourceNotFound:
                logger.warn("resource not found", resource_path=resource)
                return {}
        elif isinstance(resource, Resource):
            resource_dict = resource.dict()
        elif isinstance(resource, dict):
            resource_dict = resource
        else:
            raise ValueError("Unsupported resource type: {}".format(type(resource)))

        resource_dict['context'] = self._transform_context(resource_dict['context'])
        return resource_dict

    def is_allowed(self, user: str, action: str, resource: ResourceType, context: Context = {}) -> bool:
        """
        usage:

        authorizon.is_allowed(user, 'get', '/tasks/23')
        authorizon.is_allowed(user, 'get', '/tasks')


        authorizon.is_allowed(user, 'post', '/lists/3/todos/37', context={org_id=2})


        authorizon.is_allowed(user, 'view', task)
        authorizon.is_allowed('view', task)
        """
        resource = self._translate_resource(resource)
        if not resource:
            return False
        query_context = self._combine_context(context)
        input = {
            "user": user,
            "action": action,
            "resource": resource,
            "context": query_context,
        }
        response = requests.post(f"{SIDECAR_URL}/allowed", data=json.dumps(input))
        response_data = response.json()
        return response_data.get("allow", response_data.get("result", False))

enforcer = Enforcer()